from .logger import logger, debug, info, error, exception, level
from .net import (
    request,
    get_ip_adr,
)
from .shell import shell, bash, pipe
from .ssh import SSH, ParamSSH
from .config import get_app_config
from .tracer import Tracer



from inspect import (
    _empty,
    getargvalues,
    getouterframes,
    signature,
)

from os import getcwd
from re import compile
from sys import settrace, path
from time import perf_counter_ns


lib_paths = (
    '_bootstr',
    '<string>',
)

class Tracer:
    def __init__(self,
        label       = 'tracing',
        full_path   = False,
        show_types  = False,
        check_types = False,
        show_times  = False,
        show_stamp  = False,
        stdlib      = False,
        args_format = 'str',
        dest        = None,
        buffered    = True,
        echo        = True,
        quiet       = True,
    ):
        self.label = label
        self.level = 0
        self.full_path = full_path
        self.traces = list()
        self.show_types = show_types
        self.check_types = check_types
        self.ignored_paths = set()
        self.ignored_funcs = set()
        self.truncated = list()
        self.show_times = show_times
        self.show_stamp = show_stamp
        self.args_format = args_format
        self.dest = dest or getcwd()
        self.buffered = buffered
        self.echo = echo
        self.quiet = quiet
        self.stdlib = stdlib
        self.setup()

    @property
    def output(self):
        return f'{self.dest}/{self.label}.trace'

    def create(self):
        try:
            with open(self.output, 'x'):
                self.log(f'>>> {self.label}', ts=False)
        except FileExistsError: pass

    def write(self, line):
        with open(self.output, 'a') as out:
            out.write(f'{line}\n')

    def log(self, msg, ts=True):
        ts = f'{self.timestamp()}@{self.label[:12]:<12}' if ts and self.show_stamp else ''
        tab = '    ' * self.level
        msg = f'{ts}{tab}{msg}'
        if self.buffered:
            self.traces.append(msg)
        else:
            self.write(msg)
        if self.echo:
            print(msg)

    def log_exc(self, exc):
        if not self.quiet:
            msg = f'! {exc} !'
            self.log(msg)

    def dump(self):
        self.create()
        with open(self.output, 'a') as out:
            out.write('\n'.join(self.traces))

    def truncate(self, path):
        if not self.full_path:
            for pref in self.truncated:
                if path.startswith(pref):
                    path = path[len(pref)+1:]
        return path.removesuffix('.py')

    def ignore_path(self, path):
        return self.ignored_paths.add(compile(path).search)

    def ignored_path(self, path):
        return any(ignore(path) for ignore in self.ignored_paths)

    def ignore_func(self, func):
        return self.ignored_funcs.add(compile(func).search)

    def ignored_func(self, func):
        return any(ignore(func) for ignore in self.ignored_funcs)

    def truncate_path(self, prefix):
        self.truncated.append(prefix)

    def timestamp(self):
        return perf_counter_ns()

    def start_timer(self):
        if self.show_times:
            self.timers[self.level] = self.timestamp()

    def stop_timer(self):
        if self.show_times:
            self.timers[self.level] = self.timestamp() - self.timers[self.level]

    @property
    def elapsed(self):
        if self.show_times:
            elapsed = self.timers[self.level] // 1_000_000
            if elapsed < 10000:
                return f' [{elapsed} ms]' if elapsed > 0 else ''
            else:
                return f' [{elapsed // 1000} s]'
        else:
            return ''

    def setup(self):
        std_paths = [p for p in path if 'python' in p]
        std_paths.extend(lib_paths)
        if not self.stdlib:
            for p in std_paths:
                self.ignore_path(p)
        if not self.full_path:
            self.truncated.extend(std_paths)
        self.ignore_func(r'Tracer\..*')
        self.ignore_path('tracer')
        self.truncate_path(getcwd())
        self._exc_ = False
        self.timers = dict()

    def format_args(self, args):
        if self.args_format == 'repr':
            formatter = repr
        elif self.args_format == 'str':
            formatter = str
        else:
            raise ValueError(f'Invalid args_formatter : {self.args_format}')

        items = list()
        for k in args:
            t = type(args[k]).__name__
            try:
                v = formatter(args[k])
            except Exception as x:
                self.log_exc(x)
                v = ' ! '
            items.append((k, t, v))
        return items

    def trace(self, file, line, func, args):
        if isinstance(args, tuple) and args[0] in ('>>>', '!!!'):
            if args[0] == '>>>':
                val = args[1]
                if self.show_types:
                    args = f' -> {type(val).__name__}: {val}{self.elapsed}'
                else:
                    args = f' -> {val or ""}{self.elapsed}'
            elif args[0] == '!!!':
                args = f' ! {args[1].__name__} ! {args[2]}'
        else:
            if self.show_types:
                args = f"({', '.join(f'{k}:{t} = {v}' for k, t, v in self.format_args(args))})"
            else:
                args = f"({', '.join(f'{k}={v}' for k, t, v in self.format_args(args))})"
        self.log(f'{file}:{line} {func}{args}')

    def ann_err(self, name, expected, actual):
        if name == '_return_':
            msg = f'! "should return "{expected.__name__}" not "{actual.__name__}" !'
        else:
            msg = f'! arg "{name}" should be "{expected.__name__}" not "{actual.__name__}" !'
        self.log(msg)

    def get_func_args(self, frame):
        arginfo = getargvalues(frame)
        return {
            k:arginfo.locals[k]
            for k in arginfo.args
            if k != 'self'
        }

    def get_sig(self, frame):
        fct = frame.f_code.co_name
        if fct.endswith('comp>'): return
        slf = frame.f_locals.get('self')
        if slf:
            # method
            func = getattr(slf, fct)
        else:
            try:
                # global function
                func = frame.f_globals[fct]
            except KeyError:
                # inner function
                outer = getouterframes(frame)[1]
                func = outer.frame.f_locals[fct]
        return signature(func)
    def chk_arg_ann(self, name, arg, ann):
        if ann is _empty: return
        if ann is None and arg is None: return
        try:
            if not isinstance(arg, ann):
                self.ann_err(name, ann, type(arg))
        except TypeError:
            self.log(f'! unable to  check arg ! {arg}:{ann}')

    def check_sig(self, frame, args):
        if self.check_types:
            sig = self.get_sig(frame)
            if not sig or not sig.parameters: return
            if isinstance(args, tuple) and args[0] == '>>>':
                arg = args[1]
                ann = sig.return_annotation
                self.chk_arg_ann('_return_', arg, ann)
            else:
                if isinstance(args, dict):
                    for name, p in sig.parameters.items():
                        arg = args[name]
                        ann = p.annotation
                        if ann is _empty: continue
                        self.chk_arg_ann(name, arg, ann)

    def get_func_name(self, frame):
        fct = frame.f_code.co_name
        slf = frame.f_locals.get('self')
        try:
            repr(slf) # check handlability
        except Exception:
            slf = None
        if slf:
            try:
                func = getattr(slf, fct)
            except Exception:
                func = None
        else:
            func = frame.f_globals.get(fct)
        try:
            return func.__qualname__
        except Exception:
            return fct

    def get_trace(self, frame, arg, lvl=0):
        path = frame.f_code.co_filename
        if self.ignored_path(path): return False
        func = self.get_func_name(frame)
        if self.ignored_func(func): return False
        self.level += lvl
        self.check_sig(frame, arg)
        path = self.truncate(path)
        line = frame.f_lineno
        self.trace(path, line, func, arg)
        return path, line, func, arg

    def trace_exception(self, frame, typ, val, tbk):
        ret = self.get_trace(frame, ('!!!', typ, val))
        if ret:
            self._exc_ = True

    def trace_return(self, frame, val):
        self.stop_timer()
        if self._exc_:
            self.level-=1
            self._exc_ = False
            return
        if self.get_trace(frame, ('>>>', val)):
            self.level-=1

    def trace_call(self, frame, event, arg):
        args = self.get_func_args(frame)
        self.get_trace(frame, args, 1)
        self.start_timer()
        return self.trace_event

    def trace_event(self, frame, event, arg):
        if event == 'call':
            return self.trace_call

        if event == 'return':
            self.trace_return(frame, arg)

        if event == 'exception':
            self.trace_exception(frame, *arg)

    def __enter__(self):
        if not self.buffered:
            self.create()
        self.level = 0
        self.log(f'{self.label} >>>', ts=False)
        settrace(self.trace_event)

    def __exit__(self, typ, val, tbk):
        settrace(None)
        if self.level == 0:
            self.log(f'{self.label} <<<', ts=False)
        else:
            self.log(f'{self.label} !0!', ts=False)
        if self.buffered:
            self.dump()
        return self.quiet



from os import scandir
from pathlib import Path

from .files import grep
from .shell import bash
from .commandparams import Param, ValidationError


class Config:
    def __init__(self, ssh):
        self.ssh = ssh

    def parse(self, path='~/.ssh'):
        rex = r"\s*(?<!#)(\w+)(?:\s*=\s*|\s+)(.+)"
        hosts, host = dict(), None
        path = Path(path).expanduser()
        for cfg in scandir(path):
            if cfg.is_file and cfg.name.endswith('.cfg') and 'Host' in open(cfg).read():
                for k, v in grep(cfg, rex):
                    key = k.lower()
                    value = v.strip('"')
                    if key == "host":
                        if host:
                            hosts[host] = items
                        host = value
                        items = dict()
                    else:
                        if key in ('identityfile', 'localforward', 'remoteforward'):
                            items.setdefault(key, []).append(value)
                        else:
                            items[key] = value
                hosts[host] = items
        return hosts

    def __iter__(self):
        return iter(self.parse().items())


class Host:
    def __init__(self, ssh, name, data, shell=bash):
        self.ssh = ssh
        self.name = name
        self.data = data
        for key, val in data.items():
            setattr(self, key, val)
        self.shell = shell

    @property
    def adr(self):
        return self.hostname

    @property
    def key(self):
        return self.identityfile[0]

    def run(self, cmd, timeout=None):
        timeout = timeout or self.ssh.timeout
        return self.shell(f'''
            ssh {self.name} -o ConnectTimeout={timeout} -t {cmd}
        ''')

    def __call__(self, cmd):
        ret, out, err = self.run(cmd)
        return ret==0 and out or f'\n ! {out.strip()}\n'

    @property
    def host(self):
        return self('hostname')

    def ping(self):
        ret, out, err = self.shell(f'''
            ping -c 1 {self.hostname}
        ''')
        return ret==0

    def is_known(self):
        ret, out, err = self.shell(f'''
            ssh-keygen -F {self.hostname}
        ''')
        return ret==0

    def unknown(self):
        return self.shell(f'''
            ssh-keygen -R {self.hostname}
        ''')

    def known(self):
        if self.is_known(): return
        if not self.ping(): return
        print(f'known {self.hostname}')
        return self.shell(f'''
            touch ~/.ssh/known_hosts
            ssh-keyscan -H {self.hostname} >> ~/.ssh/known_hosts
        ''')

    def port(self, port):
        return self.shell(f'''
            ssh  -Nf -L {port}:localhost:{port} {self.name}
        ''')

    def get(self, rem, loc=None):
        return self.shell(f'''
            scp {self.name}:{rem} {loc or rem})
        ''')

    def put(self, loc, rem=None):
        return self.shell(f'''
            scp {loc} {self.name}:{rem or loc})
        ''')

    @property
    def mountpoint(self):
        return f'{self.ssh.mountpoint}/{self.name}'

    def mount(self, rem):
        loc = f'{self.mountpoint}/{rem}'
        return self.shell(f'''
            mkdir -p {loc}
            sshfs {self.name}:{rem} {loc}
        ''')

    def umount(self, loc):
        loc = f'{self.mountpoint}/{loc}'
        return self.shell(f'''
            fusermount -u {loc}
        ''')

    def sync(self, loc, rem=None, delete=True, exclude=None, dry=True):
        option = "--dry-run" if dry else "--progress"
        exclude = exclude or self.ssh.exclude
        exclude = f'--exclude-from={exclude}' if exclude else ''
        delete = '--delete' if delete else ''
        self(
            f'mkdir -p {rem}'
        )
        return self.shell(f'''
            rsync -crzhiv {exclude} {delete} {option} -e "ssh -i {self.key}" {loc} {self.user}@{self.adr}:{rem or loc}
        ''')

    def restore(self, rem, loc=None, delete=True, exclude=None, dry=True):
        option = "--dry-run" if dry else "--progress"
        exclude = exclude or self.ssh.exclude
        exclude = f'--exclude-from={exclude}' if exclude else ''
        delete = '--delete' if delete else ''
        self.shell(
            f'mkdir -p {loc}'
        )
        return self.shell(f'''
            rsync -crzhiv {exclude} {delete} {option} -e "ssh -i {self.key}" {self.user}@{self.adr}:{rem} {loc or rem}
        ''')

    def __str__(self):
        return f'\n{self.name}:\n'+'\n'.join(
            f'\t{k} = {v}' for k, v in self.data.items()
        )


class Hosts:
    def __init__(self, ssh):
        self.ssh = ssh
        self.all = dict()
        for name, items in ssh.config:
            host = Host(ssh, name, items)
            setattr(self, name, host)
            self.all[name] = host

    def known(self):
        done = set()
        for host in self:
            if host.hostname in done: continue
            host.known()
            done.add(host.hostname)

    def __iter__(self):
        return iter(self.all.values())

    def __getitem__(self, name):
        return self.all.get(name)

    def __str__(self):
        return '\n'+'\n'.join(map(str,self.all.values()))


class SSH:
    def __init__(self, **k):
        self.mountpoint = k.get('mountpoint', '/tmp/ssh')
        self.exclude = k.get('exclude')
        self.timeout = k.get('timeout', 3)
        self.config = Config(self)
        self.hosts = Hosts(self)


class ParamSSH(Param):
    '''SSH host
    '''
    name = 'SSH Host'

    def verify(self, value):
        ssh = SSH()
        host = ssh.hosts[value]
        if host:
            return host
        else:
            raise ValidationError(f'"{self.name}" unknown SSH host : {value}')



'''Generic HTTP server.

Intended to be used as a mock for testing puposes.
'''
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, HTTPServer


class RequestHandler(BaseHTTPRequestHandler):
    routes = (
        ('get', '/hello', 'hello'),
    )

    def __init__(self, *a, **k):
        self.map = dict(
            get  = dict(),
            post = dict(),
        )
        for meth, path, func in self.routes:
            self.map[meth][path] = getattr(self, func)
        self.map['post']['/shutdown'] = self.shutdown
        super().__init__(*a, **k)

    def hello(self):
        return HTTPStatus.OK, 'Hi there'

    def shutdown(self):
        setattr(self.server, '_BaseServer__shutdown_request', True)
        return 200, 'shuting down'

    def not_found(self):
        return HTTPStatus.NOT_FOUND, f'{self.command} {self.path} Not Found'

    def response(self, code, body):
        self.send_response(code)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        self.wfile.write(body.encode())

    @property
    def body(self):
        size = int(self.headers['Content-Length'])
        return self.rfile.read(size).decode()

    def do_GET(self):
        print(f'GET {self.path}\nHeaders:\n{self.headers}')
        route = self.map['get'].get(self.path, self.not_found)
        code, body = route()
        self.response(code, body)

    def do_POST(self):
        print(f'POST {self.path}\nHeaders:\n{self.headers}\nBody:\n{self.body}')
        route = self.map['post'].get(self.path, self.not_found)
        code, body = route()
        self.response(code, body)


class Server:
    def __init__(self, Handler=RequestHandler):
        self.Handler = Handler

    def serve(self, adr='localhost', port=7777):
        server = HTTPServer((adr, port), self.Handler)
        print(f'Listening on {adr}:{port} ...\n')
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            pass
        server.server_close()



from sys import exit
from types import FunctionType

from pybrary.command import parse_args
from pybrary import commands


def main():
    a, k = parse_args()

    try:
        f, *a = a
    except ValueError:
        f = ''

    try:
        cmd = getattr(commands, f)
    except AttributeError:
        print('Available Pybrary Commands :')
        for k, v in vars(commands).items():
            if isinstance(v, FunctionType):
                print('   ', k)
        exit()

    try:
        success, output = cmd(*a, **k)
    except Exception as x:
        success, output = False, str(x)

    if success:
        print(output)
    else:
        exit(f' ! {output}')



from inspect import currentframe, getouterframes
from operator import attrgetter
from pathlib import Path


# Full Qualified Name od obj
def fqn(obj):
    mod = getattr(obj, '__module__', None)
    cls = obj.__class__.__name__
    if 'builtin' in cls:
        mod = cls = None
    elif cls in ('type', 'function', 'method'):
        cls = None
    name = getattr(obj, '__name__', None)
    return '.'.join(n for n in (mod, cls, name) if n)


# caller's caller
def caller():
    frame = getouterframes(currentframe(), 2)[2]
    mod = Path(frame.filename).stem
    fct = frame.function
    return f'{mod}.{fct}'


# Default implementation for abstract methods
def todo(self):
    cls = self.__class__.__name__
    mth = getouterframes(currentframe())[1][3]
    raise NotImplementedError(
        f'\n ! {mth} missing in {cls} !\n'
    )


# Memoizing property decorator
def memo(fct):
    name = '_memo_' + fct.__name__
    get_val = attrgetter(name)
    @property
    def wrapper(self):
        try:
            return get_val(self)
        except AttributeError:
            val = fct(self)
            setattr(self, name, val)
            return val
    return wrapper





from concurrent.futures import ThreadPoolExecutor, as_completed
from ipaddress import ip_address, IPv4Address
from json import loads
from random import shuffle
from urllib.request import urlopen

from .commandparams import Param, ValidationError
from . import debug, error


ip_adr_getters = [
    'api.ipify.org',
    'api.my-ip.io/ip',
    'ident.me',
]


def request(url, json=False, timeout=2):
    try:
        with urlopen(url, timeout=timeout) as resp:
            if resp.status == 200:
                txt = resp.read().decode().strip()
                if json:
                    try:
                        return True, loads(txt)
                    except Exception as x:
                        reason = f'{type(x)} {x}'
                else:
                    return True, txt
            else:
                reason = f'{resp.status} {resp.reason}'
    except Exception as x:
        reason = f'{type(x)} {x}'

    reason = f'{url} ! {reason}'
    return False, reason


def get_text(url):
    return request(f'https://{url}')


def get_ip_adr():
    with ThreadPoolExecutor(
        max_workers = len(ip_adr_getters)
    ) as ex:
        for ok, result in ex.map(
            get_text,
            ip_adr_getters,
        ):
            if ok:
                return True, result
    error('unabled to get ip adr')
    return False, 'ERROR'


class ParamIPv4(Param):
    name = 'IPv4 adr'

    def verify(self, value):
        try:
            adr = ip_address(value)
            if not isinstance(adr, IPv4Address):
                raise ValueError
            return adr
        except ValueError:
            raise ValidationError(f'{value} is not IPv4')




from functools import partial
from re import compile as rec


# remove ansi codes
rm_ansi_codes = partial(rec(
    r'\x1B\[([0-9]{1,3}(;[0-9]{1,2})?)?[mGKJH](\x0F)?'
).sub, '')


# make valid name
clean_name_re = partial(rec(r'\W').sub, '_')

def clean_name(txt):
    return clean_name_re(txt).strip('_')


# Convert file permissions from str to oct
def oct_mod(str_mod):
    assert len(str_mod)==9, f'bad mode : {str_mod} ({len(str_mod)})'

    def digit(str_bits):
        assert len(str_bits)==3, f'bad bits : {str_bits}'
        value = sum(
            2**i if c!='-' else 0
            for i, c in enumerate(str_bits[::-1])
        )
        return str(value)

    digits = [
        digit(str_mod[i*3:i*3+3])
        for i in range(3)
    ]

    return  ''.join(digits)




from functools import partial
import shlex
from subprocess import (
    CalledProcessError,
    PIPE,
    Popen,
    run,
)

from pybrary.ascii import rm_ansi_codes


def shell(script, shell, color=False):
    try:
        proc = run(
            script,
            stdout = PIPE,
            stderr = PIPE,
            check  = True,
            shell  = True,
            executable = shell,
        )
    except CalledProcessError as exc:
        ret = exc.returncode
        out = exc.stdout.decode('utf-8')
        err = exc.stderr.decode('utf-8')
    else:
        ret = proc.returncode
        out = proc.stdout.decode('utf-8').strip()
        err = proc.stderr.decode('utf-8').strip()

    if not color:
        out = rm_ansi_codes(out)
        err = rm_ansi_codes(err)

    return ret, out, err


def pipe(*cmds):
    prev = None
    for cmd in cmds:
        cmd = shlex.split(cmd)
        if prev:
            proc = Popen(cmd, stdin=prev.stdout, stdout=PIPE)
        else:
            proc = Popen(cmd, stdout=PIPE)
        prev = proc

    for line in proc.stdout.read().decode().strip().split('\n'):
        yield line.strip()


bash = partial(shell, shell='/bin/bash')


